(defpackage klee-ktest-generator
  (:use :cl :alexandria :serapeum)
  (:documentation "Tools for generating files in KLEE's ktest format."))

(in-package klee-ktest-generator)

(defparameter *ktest-magic* (map 'vector #'char-code "KTEST"))
(defparameter *ktest-version* 3)
(defparameter *ktest-model-version* #(1 0 0 0))

(defparameter *fake-stdin-stat*
  #(26 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 164 129 0 0 232 3 0
    0 232 3 0 0 1 1 1 1 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 0 0 0 0 0 0 0 0 1
    1 1 1 1 1 1 1 159 123 72 92 0 0 0 0 1 1 1 1 1 1 1 1 11 126 72 92 0 0
    0 0 1 1 1 1 1 1 1 1 11 126 72 92 0 0 0 0 1 1 1 1 1 1 1 1 1 1 1 1 1 1
    1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1))

(defun write-magic (s)
  "Write the ktest magic bytes to S."
  (write-sequence *ktest-magic* s))

(defun write-integer (n s)
  "Write N to S as a 32-bit big-endian value."
  (check-type n (integer 0 4294967295) "a 32-bit unsigned integer")
  (write-sequence (pad-start (octets n :big-endian t) 4 0) s))

(defun write-ktest-string (string s)
  "Write first the length and then each byte of STRING to S."
  (write-integer (length string) s)
  (do-each (char string)
    (write-byte (char-code char) s)))

(defun write-object (object s)
  "Write OBJECT, which must be of the form (NAME . DATA), to S."
  (destructuring-bind (name . data) object
    (write-ktest-string name s)
    (write-integer (length data) s)
    (write-sequence data s)))

(defun generate (stream program-args objects)
  "Generate a ktest file with the given OBJECTS and write it to STREAM."
  (write-magic stream)
  (write-integer *ktest-version* stream)
  (write-integer (length program-args) stream)
  (dolist (arg program-args)
    (write-ktest-string arg stream))

  ;; We don't support symbolic arguments
  (write-integer 0 stream)
  (write-integer 0 stream)

  (write-integer (1+ (length objects)) stream)
  (dolist (obj objects)
    (write-object obj stream))
  (write-object (cons "model_version" *ktest-model-version*) stream))

(defun generate-with-stdin (stream program data &rest arguments)
  "Generate a ktest file with DATA for stdin and write it to STREAM."
  (let ((stdin-length (length data)))
    (generate stream
              (append (list program "--sym-stdin" (fmt "~d" stdin-length))
                      arguments)
              (list (cons "stdin" data)
                    (cons "stdin-stat" *fake-stdin-stat*))))
  (values))

(example
  ;; (flex:with-output-to-sequence (out)
  ;;   (generate-with-stdin out "cat.bc" #(42) "--sym-stdin" "7"))
  (with-open-file (out "/tmp/generated_ktest" :direction :output
                                              :element-type '(unsigned-byte 8)
                                              :if-exists :supersede)
    (generate-with-stdin out "cat.bc" (map 'vector #'char-code "Hello!"))))

;;; Creating ktest files from all PoVs:
;; (loop for file in (cgc-pov-converter:all-povs)
;;       do (let ((conv-pov (concat (namestring file) ".bin2"))
;;                (ktest-name (concat (namestring file) ".ktest")))
;;            (when (uiop:file-exists-p conv-pov)
;;              (with-output-to-file (out ktest-name
;;                                        :element-type '(unsigned-byte 8)
;;                                        :if-exists :supersede)
;;                (generate-with-stdin out "ignored" (read-file-into-byte-vector conv-pov))))))
